#!/usr/bin/perl -w
###############################################################################
##                                                                           ##
##   A P A L O G F I L T E R . P L                                           ##
##                                                                           ##
## ------------------------------------------------------------------------- ##
##                                                                           ##
##  Job .......: Script to filter and transform an existing apache-logfile.  ##
##                                                                           ##
## ------------------------------------------------------------------------- ##
##                                                                           ##
##  Copyright (C) 2000  Daniel Scheibli                                      ##
##                                                                           ##
##  This program is free software; you can redistribute it and/or modify it  ##
##  under the terms of the GNU General Public License as published by the    ##
##  Free Software Foundation; either version 2 of the License, or (at your   ##
##  option) any later version.                                               ##
##                                                                           ##
##  This program is distributed in the hope that it will be useful, but      ##
##  WITHOUT ANY WARRANTY; without even the implied warranty of               ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            ##
##  GNU General Public License for more details.                             ##
##                                                                           ##
##  You should have received a copy of the GNU General Public License along  ##
##  with this program; if not, write to the Free Software Foundation, Inc.,  ##
##  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.                 ##
##                                                                           ##
## ------------------------------------------------------------------------- ##
##                                                                           ##
##  Author .....: Daniel Scheibli <daniel@scheibli.com>                      ##
##  Date .......: 2000-05-17                                                 ##
##  Changes ....: ----- ApaLogFilter 0.99.010 ------------------------------ ##
##                2006-09-03 o Updated code layout and program outputs.      ##
##                ----- ApaLogFilter 0.99.009 ------------------------------ ##
##                2002-08-06 o Improved performance by differenciating the   ##
##                             RequestUrl.n.Value content into cases with    ##
##                             STRING and with REGEXP value.                 ##
##                ----- ApaLogFilter 0.99.008 ------------------------------ ##
##                2001-03-12 o Added prototype for function main().          ##
##                2001-09-18 o FEATURE: Added new keyword $ALL for           ##
##                             RequestUrl.n.Title                            ##
##                ----- ApaLogFilter 0.99.007 ------------------------------ ##
##                2001-01-22 o FEATURE: Now possible, to use RegExp's for    ##
##                             Remote Host field.                            ##
##                ----- ApaLogFilter 0.99.006 ------------------------------ ##
##                2000-12-30 o FEATURE: Added feature to allow all records   ##
##                             to go to the outfile when there are no        ##
##                             RegExp. entry (for matching) are defined      ##
##                             at all.                                       ##
##                2000-12-31 o Changed all comments in code to english.      ##
##                2000-12-31 o Changed the documentation.                    ##
##                ----- ApaLogFilter 0.99.005 ------------------------------ ##
##                2000-12-29 o BUGFIX: Negative TZ not gets accepted.        ##
##                ----- ApaLogFilter 0.99.004 ------------------------------ ##
##                2000-12-03 o Changed the license to GPL.                   ##
##                2000-12-03 o Changed the documentation.                    ##
##                ----- ApaLogFilter 0.99.003 ------------------------------ ##
##                2000-08-14 o FEATURE: Added the function of replacing      ##
##                             HTTP statuscodes.                             ##
##                ----- ApaLogFilter 0.99.002 ------------------------------ ##
##                2000-06-12 o Changed comments in code.                     ##
##                2000-06-13 o FEATURE: Added the function of a seperate     ##
##                             ExceptionFile.                                ##
##                                                                           ##
###############################################################################



# Presettings

use English;
use strict;

use File::Basename;
use FindBin qw( $Bin );
use IO::File;

sub TRUE    { 1 }
sub FALSE   { 0 }
sub NULL    { undef }   # DANGER: Never change this!!!
sub OK      { 0 }
sub WARNING { 2 }
sub ERROR   { 4 }

sub Program_Init_ParseArguments();
sub Program_Init_ConfigRead();
sub Program_Action_ProcessAccessLog();

my %hGlobalVars = (
	"Program.Rc"                      => 0,                         # Return-Code
	"Program.Version"                 => "APALOGFILTER 0.99.010",   # Version-Token
	"Program.Environment.NewLine"     => "\n",                      # NewLine-Token for the current environment
	"Program.Environment.Seperator"   => "/",                       # PathSeperator-Token for the current environment
	"Program.Obj.CfgFile.Name"        => "ApaLogFilter.cfg",        # Name of the Configurationfile
	"Program.Param.InputFileName"     => "",                        # Name of the Apache-Logfile to process (INPUT)
	"Program.Param.OutputFileName"    => "",                        # Name of the Apache-Logfile for the results (OUTPUT)
	"Program.Param.ExceptionFileName" => "",                        # Name of the Exception-Logfile for Exception-Rows
	"Program.Config.RemoteHost"       => [],                        # Settings for the "Client"-field
	"Program.Config.RequestUrl"       => [],                        # Settings for the "URL"-field
	"Program.Config.Status"           => [],                        # Settings for the "Status"-field
	"Program.Config.ReferredUrl"      => []                         # Settings for the "ReferredURL"-field
);
$hGlobalVars{"Program.Obj.CfgFile.Name"} = $Bin.
	$hGlobalVars{"Program.Environment.Seperator"}.
	$hGlobalVars{"Program.Obj.CfgFile.Name"};



# Program initialization

printf( "%s%s", 
	$hGlobalVars{"Program.Version"}, 
	$hGlobalVars{"Program.Environment.NewLine"} );

Program_Init_ParseArguments();
Program_Init_ConfigRead();

printf( "INFO: Number of RequestUrl entries = %s%s", 
	scalar@{$hGlobalVars{"Program.Config.RequestUrl"}}, 
	$hGlobalVars{"Program.Environment.NewLine"} );



# Process the web server's access log file

Program_Action_ProcessAccessLog();

exit( OK );





###############################################################################
##                                                                           ##
##   P R O G R A M _ I N I T _ P A R S E A R G U M E N T S                   ##
##                                                                           ##
###############################################################################
sub Program_Init_ParseArguments() {
	
	# Presettings

	my( $sCallRc );



	# Process the parameters/arguments given to the program

	my @aParamMarkerDefinitions = (
		{
			"HashKey"    => "Program.Param.InputFileName",
			"ArgToken"   => "-in",
			"Definition" => "STRING:Y"
		},{
			"HashKey"    => "Program.Param.OutputFileName",
			"ArgToken"   => "-out",
			"Definition" => "STRING:Y"
		},{
			"HashKey"    => "Program.Param.ExceptionFileName",
			"ArgToken"   => "-exf",
			"Definition" => "STRING:Y"
		}			  
	);
	
	$sCallRc = ParseArguments( \@ARGV, \@aParamMarkerDefinitions, \%hGlobalVars );

	if( $sCallRc != OK ) {
	
		my( @aRows );
		
		push( @aRows, "===> ERROR: Unknown arguments (Rc: ".$sCallRc.")." );
		push( @aRows, "" );
		push( @aRows, "Call ........: >>--- ApaLogFilter.pl --- -in FILE --- -out FILE --->" );
		push( @aRows, "" );
		push( @aRows, "                >--- -exf FILE ---><" );
		push( @aRows, "" );
		push( @aRows, "Sample ......: ApaLogFilter.pl -in access.log -out access.flg -exf access.err" );
		push( @aRows, "" );
		push( @aRows, "Arguments ...: -in FILE    Full qualified name of the AccessLogFile" );
		push( @aRows, "                           to process." );
		push( @aRows, "               -out FILE   Full qualified name of the ResultFile." );
		push( @aRows, "               -exf FILE   Full qualified name of the ExceptionFile." );  
		push( @aRows, "" );
		
		foreach( @aRows ) { 
			printf( "%s%s", $_, $hGlobalVars{"Program.Environment.NewLine"} );
		}

		exit( ERROR );		
	}
}





###############################################################################
##                                                                           ##
##   P R O G R A M _ I N I T _ C O N F I G R E A D                           ##
##                                                                           ##
###############################################################################
sub Program_Init_ConfigRead() {

 	# Presettings

	my( $oCfgFile, $rhKeyValue, $sIndex, $sSubIndex );
	my( $rhReqExpArray, $rhStringArray );
	my( $sCallRc, $sDummy, $I );



	# Open and read the ConfigFile

	$rhKeyValue = ParseConfigurationFile( $hGlobalVars{"Program.Obj.CfgFile.Name"} );
	if( ! ref( $rhKeyValue ) ) {
		printf( "===> ERROR: Unable to open/read the ConfigFile (Rc: %s)%s", $sCallRc, $hGlobalVars{"Program.Environment.NewLine"} );
		exit( ERROR );
	}



	# Process the settings for the "RemoteHost"-field

	for( $sIndex = 1; ; $sIndex++ ) {
		
		# If there is a RemoteHost.<n> Entry,...
		
		if( defined( $rhKeyValue->{"REMOTEHOST.".$sIndex.".VALUE"} ) ) {

			# Ensure, that the settings (RemoteHost.<n>.Value) are ok
				
			if( $rhKeyValue->{"REMOTEHOST.".$sIndex.".VALUE"} eq "" ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} );
				printf( "            The value of REMOTEHOST.%s.VALUE is a nullstring.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}

			# Ensure, that the settings (RemoteHost.<n>.Action) are ok
				
			unless( ( defined( $rhKeyValue->{"REMOTEHOST.".$sIndex.".ACTION"} ) ) &&
				( $rhKeyValue->{"REMOTEHOST.".$sIndex.".ACTION"} eq "EXCLUDE_RECORD" ) ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} );
				printf( "            The value of REMOTEHOST.%s.ACTION is invalid or undefined.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}
 	
			# Added the settings
				
			push( @{$hGlobalVars{"Program.Config.RemoteHost"}},
				{
					"VALUE"  => $rhKeyValue->{"REMOTEHOST.".$sIndex.".VALUE"},
					"ACTION" => $rhKeyValue->{"REMOTEHOST.".$sIndex.".ACTION"}
				}
			);
		}
		else {
			last;
		}
	}  



	# Process the settings for the "RequestUrl"-field

	for( $sIndex = 1; ; $sIndex++ ) {
		
		# If there is a RequestUrl.<n> Entry,...
			
		if( defined( $rhKeyValue->{"REQUESTURL.".$sIndex.".TITLE"} ) ) {
			
			# Ensure, that the settings (RequestUrl.<n>.Title) are ok

			if( $rhKeyValue->{"REQUESTURL.".$sIndex.".TITLE"} eq "" ) {
				printf( "===> ERRROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
				printf( "             The value of REQUESTURL.%s.TITLE is a nullstring.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}

			# Walk thrue all possible URL's / ReqExp's
				
			$rhReqExpArray = [];
			$rhStringArray = [];
				
			for( $sSubIndex = 1; ; $sSubIndex++ ) {
					
				# If there is a RequestUrl.<n>.Value.<m> Entry,...
					
				if( ( defined( $rhKeyValue->{"REQUESTURL.".$sIndex.".VALUE.".$sSubIndex} ) ) &&
					( defined( $rhKeyValue->{"REQUESTURL.".$sIndex.".TYPE.".$sSubIndex} ) ) ) {
						
					# Ensure, that the settings (RequestUrl.<n>.Value.<m>) are ok
						
					if( $rhKeyValue->{"REQUESTURL.".$sIndex.".VALUE.".$sSubIndex} eq "" ) {
						printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
						printf( "            The value of REQUESTURL.%s.VALUE.%s is invalid or undefined.%s", $sIndex, $sSubIndex, $hGlobalVars{"Program.Environment.NewLine"} );
						exit( ERROR );
					}
						
					# Ensure, that the settings (RequestUrl.<n>.Type.<m>) are ok     
						
					if( $rhKeyValue->{"REQUESTURL.".$sIndex.".TYPE.".$sSubIndex} eq "" ) {
						printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
						printf( "            The value of REQUESTURL.%s.TYPE.%s is invalid or undefined.%s", $sIndex, $sSubIndex, $hGlobalVars{"Program.Environment.NewLine"} );
						exit( ERROR );
					}

					# Added the settings
					
					if( $rhKeyValue->{"REQUESTURL.".$sIndex.".TYPE.".$sSubIndex} eq "STRING" ) {
						push( @{$rhStringArray}, $rhKeyValue->{"REQUESTURL.".$sIndex.".VALUE.".$sSubIndex} );
					}
					elsif( $rhKeyValue->{"REQUESTURL.".$sIndex.".TYPE.".$sSubIndex} eq "REGEXP" ) {
						push( @{$rhReqExpArray}, $rhKeyValue->{"REQUESTURL.".$sIndex.".VALUE.".$sSubIndex} );
					}
					else {
						printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
						printf( "            The value of REQUESTURL.%s.TYPE.%s is invalid or undefined.%s", $sIndex, $sSubIndex, $hGlobalVars{"Program.Environment.NewLine"} );
						exit( ERROR );
					}      
				}
				else {
					last;
				}
			}
    
			# Ensure, that there is a minimum of one URL / ReqExp per RequestUrl.<n>.Title entry
				
			if( ( scalar@{$rhStringArray} < 1 ) && ( scalar@{$rhReqExpArray} < 1) ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} );
				printf( "            No entry for REQUESTURL.%s.VALUE.<n> found.%s", $sIndex, $sSubIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}
    
			# Added the settings
			
			push( @{$hGlobalVars{"Program.Config.RequestUrl"}},
				{
					"TITLE"  => $rhKeyValue->{"REQUESTURL.".$sIndex.".TITLE"},
					"STRING" => $rhStringArray,
					"REGEXP" => $rhReqExpArray
				}
			);    
		}
		else {
			last;
		}
	}  



	# Process the settings for the "Status"-field

	for( $sIndex = 1; ; $sIndex++ ) {

		# If there is a Status.<n> Entry,...
			
		if( defined( $rhKeyValue->{"STATUS.".$sIndex.".VALUE"} ) ) {
			
			# Ensure, that the settings (Status.<n>.Value) are ok
				
			if( $rhKeyValue->{"STATUS.".$sIndex.".VALUE"} eq "" ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} );
				printf( "            The value of STATUS.%s.VALUE is a nullstring.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}

			# Ensure, that the settings (Status.<n>.Action) are ok
				
			unless( ( defined( $rhKeyValue->{"STATUS.".$sIndex.".ACTION"} ) ) &&
				( ( $rhKeyValue->{"STATUS.".$sIndex.".ACTION"} eq "FORCE_RECORD_IN_EXCEPTION_LIST" ) ||
				  ( $rhKeyValue->{"STATUS.".$sIndex.".ACTION"} eq "REPLACE_VALUE" ) ) ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
				printf( "            The value of STATUS.%s.ACTION is invalid or undefined.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}

			# Ensure, that the there is an NewValue for the Action "Replace_Value"
				
			if( $rhKeyValue->{"STATUS.".$sIndex.".ACTION"} eq "REPLACE_VALUE" ) {
				unless( ( defined( $rhKeyValue->{"STATUS.".$sIndex.".NEWVALUE"} ) ) &&
					( $rhKeyValue->{"STATUS.".$sIndex.".NEWVALUE"} ne "" ) ) {
					printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
					printf( "            The value of STATUS.%s.NEWVALUE is invalid or undefined.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
					exit( ERROR );
				}
			}
 	
			# Added the settings
				
			push( @{$hGlobalVars{"Program.Config.Status"}},
				{
					"VALUE"    => $rhKeyValue->{"STATUS.".$sIndex.".VALUE"},
					"ACTION"   => $rhKeyValue->{"STATUS.".$sIndex.".ACTION"},
					"NEWVALUE" => $rhKeyValue->{"STATUS.".$sIndex.".NEWVALUE"}
				}
			);
		}
		else {
			last;
		}
	}  



	# Process the settings for the "ReferredUrl"-field

	for( $sIndex = 1; ; $sIndex++ ) {

		# If there is a ReferredUrl.<n> Entry,...
			
		if( defined( $rhKeyValue->{"REFERREDURL.".$sIndex.".VALUE"} ) ) {

			# Ensure, that the settings (ReferredUrl.<n>.Value) are ok
				
			if( $rhKeyValue->{"REFERREDURL.".$sIndex.".VALUE"} eq "" ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
				printf( "            The value of REFERREDURL.%s.VALUE is a nullstring.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );      
				exit( ERROR );
			}

			# Ensure, that the settings (ReferredUrl.<n>.Action) are ok

			unless( ( defined( $rhKeyValue->{"REFERREDURL.".$sIndex.".ACTION"} ) ) &&
				( $rhKeyValue->{"REFERREDURL.".$sIndex.".ACTION"} eq "REPLACE_WITH_BLANK" ) ) {
				printf( "===> ERROR: The ConfigFile contains an invalid entry.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
				printf( "            The value of REFERREDURL.%s.ACTION is invalid or undefined.%s", $sIndex, $hGlobalVars{"Program.Environment.NewLine"} );
				exit( ERROR );
			}
 	
			# Added the settings
				
			push( @{$hGlobalVars{"Program.Config.ReferredUrl"}},
				{
					"VALUE"  => $rhKeyValue->{"REFERREDURL.".$sIndex.".VALUE"},
					"ACTION" => $rhKeyValue->{"REFERREDURL.".$sIndex.".ACTION"}
				}
			);
		}
		else {
			last;
		}
	}  
}





###############################################################################
##                                                                           ##
##   P R O G R A M _ A C T I O N _ P R O C E S S A C C E S S L O G           ##
##                                                                           ##
###############################################################################
sub Program_Action_ProcessAccessLog() {

	# Presettings
	
	my(
		$sClient,   $sIdentUser, $sAuthUser,
		$sDate,     $sTime,      $sTz,
		$sMethod,   $sUrl,       $sProtocol,
		$sStatus,   $sBytes,
		$sReferred, $sBrowser
	);
	my( $sRowNo, $sRow );    
	my( $sFound );
	my( $oFhOut, $oFhException, $oFhIn, $oFhDummy );
	my( $sCallRc, $sDummy, $I, $J, @aDummy );



	# Open output file

	$sDummy = 0;
	$oFhOut = new IO::File;
	$oFhOut->open( $hGlobalVars{"Program.Param.OutputFileName"}, O_WRONLY | O_CREAT | O_TRUNC ) or $sDummy = 1; 
	
	if( $sDummy != 0 ) { 
		printf( "===> ERROR: Unable to open the ResultFile.%s", $hGlobalVars{"Program.Environment.NewLine"} );
		exit( ERROR );       
	}
	$oFhOut->autoflush( 1 );
 


	# Open exception file
		
	$sDummy = 0;
	$oFhException = new IO::File;
	$oFhException->open( $hGlobalVars{"Program.Param.ExceptionFileName"}, O_WRONLY | O_CREAT | O_TRUNC ) or $sDummy = 1; 
		
	if( $sDummy != 0 ) {
		printf( "===> ERROR: Unable to open the ExceptionFile.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
		exit( ERROR );
	}
	$oFhException->autoflush(1);


 
	# Open the web server access log file

	$sDummy = 0;
	$oFhIn  = new IO::File;
	$oFhIn->open( $hGlobalVars{"Program.Param.InputFileName"}, O_RDONLY ) or $sDummy = 1;

	if( $sDummy != 0 ) {
		printf( "===> ERROR: Unable to open the AccessLogFile.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
		exit( ERROR );
	}



	# Walk through the web server access log files
		
	READ_LOOP: for(;;) {
		
		# Presettings
		
		$sRowNo++;
		if( $sRowNo % 500 == 0 ) {
			printf( "Processing row number %ld ...%s", $sRowNo, $hGlobalVars{"Program.Environment.NewLine"} );
		}
		$oFhDummy = $oFhOut;

		# Read next row
			
		$_ = $oFhIn->getline();
		if( defined( $_ ) ) {
			chomp( $_ );
		}
		else {
			last READ_LOOP;
		}

		# Split of the aktual row into the contained tokens/fields
			
		( $sClient,   $sIdentUser, $sAuthUser,
			$sDate,     $sTime,      $sTz,
			$sMethod,   $sUrl,       $sProtocol,
			$sStatus,   $sBytes,
			$sReferred, $sBrowser ) = /^(\S+) (\S+) (\S+) \[(\S+):(\d+:\d+:\d+) (\S+)\] "(\S+) (.*?) (\S+)" (\S+) (\S+) "(.*?)" "(.*?)"/;

		# Ensure, that all fields could be reed
			
		if( ! defined( $sBrowser ) ) {
			printf( "===> ERROR: Corrupted Record (RowNo: %i).%s", $sRowNo, $hGlobalVars{"Program.Environment.NewLine"} );
			exit( ERROR );
		}

		ANALYSE_LOOP: for(;;) {
			
			# Walk thrue the settings for status field
				
			for( $I = 0; $I < scalar@{$hGlobalVars{"Program.Config.Status"}}; $I++ ) {
				
				# If status matches,...
					
				if( $hGlobalVars{"Program.Config.Status"}->[$I]->{"VALUE"} eq $sStatus ) {
					
					# If transfer of row to exception file is wanted,...
						
					if( $hGlobalVars{"Program.Config.Status"}->[$I]->{"ACTION"} eq "FORCE_RECORD_IN_EXCEPTION_LIST" ) {
						$oFhDummy = $oFhException;
						last ANALYSE_LOOP;
					}
						
					# If replacement of the value is wanted,...
						
					if( $hGlobalVars{"Program.Config.Status"}->[$I]->{"ACTION"} eq "REPLACE_VALUE" ) {
						$sStatus = $hGlobalVars{"Program.Config.Status"}->[$I]->{"NEWVALUE"};
					}
				}      
			}   

   
			# Walk thrue the settings for remote host field (only for EXCLUDE)
				
			for( $I = 0; $I < scalar@{$hGlobalVars{"Program.Config.RemoteHost"}}; $I++ ) {
				
				# Presettings
						
				$sDummy = $hGlobalVars{"Program.Config.RemoteHost"}->[$I]->{"VALUE"};
						
				# If remote host matches,...
					
				if( $sClient =~ /^$sDummy$/ ) {
					
					# If EXCLUDE is wanted,...
						
					if( $hGlobalVars{"Program.Config.RemoteHost"}->[$I]->{"ACTION"} eq "EXCLUDE_RECORD" ) {
						next READ_LOOP;
					}
				}
			}

			# Walk thrue the settings for request url
				
			$sFound = 0;
			LOOP3: for( $I = 0; $I < scalar@{$hGlobalVars{"Program.Config.RequestUrl"}}; $I++ ) {
				for( $J = 0; $J < scalar@{$hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"STRING"}}; $J++ ) {
				
					# Presettings
						
					$sDummy = $hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"STRING"}->[$J];
							
					# If request url matches,...
							
					if( $sUrl eq $sDummy ) {
							
						$sFound = 1;
						$sDummy = Translate( ' ', '_', undef, $hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"TITLE"} );
						if( index( $sDummy, '$ALL' ) >= 0 ) {
							$sDummy = $sUrl;
						}
						elsif( index( $sDummy, '$FILE' ) >= 0 ) {
							$sDummy = Replace( '\$FILE', basename( $sUrl ), $sDummy );
						}
						$sUrl = $sDummy;
						last LOOP3;
					}
				}
				for( $J = 0; $J < scalar@{$hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"REGEXP"}}; $J++ ) {
					
					# Presettings
						
					$sDummy = $hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"REGEXP"}->[$J];
					
					# If request url matches,...
						
					if( $sUrl =~ /^$sDummy$/ ) {
						
						$sFound = 1;
						$sDummy = Translate( ' ', '_', undef, $hGlobalVars{"Program.Config.RequestUrl"}->[$I]->{"TITLE"} );
						if( index( $sDummy, '$ALL' ) >= 0 ) {
							$sDummy = $sUrl;
						}
						elsif( index( $sDummy, '$FILE' ) >= 0 ) {
							$sDummy = Replace( '\$FILE', basename( $sUrl ), $sDummy );
						}
						$sUrl = $sDummy;
						last LOOP3;
					}
				}
			}

			# If current row doesn't match any defined Request-URL,...

			if( ( $sFound == 0 ) && ( scalar@{$hGlobalVars{"Program.Config.RequestUrl"}} > 0 ) ) {
				next READ_LOOP;
			}

			# Walk thrue the settings for referred url (for REPLACE)
				
			for( $I = 0; $I < scalar@{$hGlobalVars{"Program.Config.ReferredUrl"}}; $I++ ) {
				
				# If referred url matches and replacement is wanted,...
					
				if( ( index( $sReferred, $hGlobalVars{"Program.Config.ReferredUrl"}->[$I]->{"VALUE"} ) == 0 ) &&
					( $hGlobalVars{"Program.Config.ReferredUrl"}->[$I]->{"ACTION"} eq "REPLACE_WITH_BLANK" ) ) {
					$sReferred = "-";
				}
			}
				
			last ANALYSE_LOOP;
		}

		# Assemble the new row in CLF format
			
		$sRow = $sClient;
		$sRow = $sRow." ".  $sIdentUser;
		$sRow = $sRow." ".  $sAuthUser;
		$sRow = $sRow." [". $sDate;
		$sRow = $sRow.":".  $sTime;
		$sRow = $sRow." ".  $sTz;
		$sRow = $sRow.'] "'.$sMethod;
		$sRow = $sRow." ".  $sUrl;
		$sRow = $sRow." ".  $sProtocol;   
		$sRow = $sRow.'" '. $sStatus;   
		$sRow = $sRow." ".  $sBytes;  
		$sRow = $sRow.' "'. $sReferred;
		$sRow = $sRow.'" "'.$sBrowser;
		$sRow = $sRow.'"';   

		# Write new row to output / exception file
			
		$sDummy = 0;
		$oFhDummy->printf( "%s%s", $sRow, $hGlobalVars{"Program.Environment.NewLine"} ) or $sDummy = 1;
			
		if( $sDummy != 0 ) {
			printf( "===> ERROR: Unable to write into the Result/ExceptionFile.%s", $hGlobalVars{"Program.Environment.NewLine"} ); 
			exit( ERROR );
		}   
	}
		
	close($oFhIn);
	close($oFhException);  
	close($oFhOut);
}





###############################################################################
##                                                                           ##
##   P A R S E C O N F I G U R A T I O N F I L E                             ##
##                                                                           ##
###############################################################################
sub ParseConfigurationFile($) {

	# NOTE: This  code is very basic, but works and allows one to
	# parse a file like the one described in the box (the standard
	# Perl functions are not sufficient for that).

	# TODO: Have a second look at the available CPAN modules.
	
	# +------------------------------------------------------------------+
	# |  #BEGIN                                                          |
	# |  # CommentToken                                                  |
	# |  KeyToken1.KeyToken2 = ValueToken1 ValueToken2 \ # CommentToken  |
	# |                        ValueToken3               # CommentToken  |
	# |  #END                                                            |
	# +------------------------------------------------------------------+



	# Presettings

	my( $sFileName ) = @_;	
	my $oFh = undef;
	my $rhKeyValueRows = {};
	my $sDummy;



	# Open the configuration file

	$sDummy = OK;
	$oFh = new IO::File;
	$oFh->open( $sFileName, O_RDONLY ) or $sDummy = ERROR;
	
	if( $sDummy != OK ) {
		$oFh = undef;
		return( -5 );
	}


 
	# Walk thrue all rows in config file

	my( $sBeginRowFlag, $sEndRowFlag, $sRowNo ) = ( FALSE, FALSE, 0 );
	my( $sCurrentRowOrg );
	my( $sCurrentRowKeyValue, $sCurrentRowComment );
	my( $sCurrentRowKey, $sCurrentRowValue );
	my( $sCommentPos, $sEqualPos );
	my( $I, $sBackSlashesNoOf );
	my( $sKey4FollowRow, $sRowNo2Follow ) = ( "", -1 );
	
	while( defined( $sCurrentRowOrg = $oFh->getline() ) ) {
				
		# Presettings
			
		chomp( $sCurrentRowOrg );
		$sCommentPos = index( $sCurrentRowOrg, '#' );
		$sRowNo++;   
   
		# Cut out the comment from row
			
		if( $sCommentPos < 0 ) {
			$sCommentPos = length( $sCurrentRowOrg );
		}
		$sCurrentRowKeyValue = Strip( substr( $sCurrentRowOrg, 0, $sCommentPos ) );
		$sCurrentRowComment  = Strip( Space( 0, ' ', uc( substr( $sCurrentRowOrg, $sCommentPos ) ) ) );

		# Check for pure comment row
			
		if( ( $sCurrentRowKeyValue eq "" ) && ( $sCurrentRowComment ne "" ) ) {
			$sRowNo--;
			next;
		}
			
		# Check for key/value or value containing rows
			
		elsif( $sCurrentRowKeyValue ne "" ) {
			
			# Presettings
				
			$sEqualPos = index( $sCurrentRowKeyValue, '=' );
			
			# If row is a key/value row,...
			
			if( $sEqualPos >= 0 ) {
				
				# Split the row into key- and value-part
				
				$sCurrentRowKey   = Strip( uc( substr( $sCurrentRowKeyValue, 0, $sEqualPos ) ) );   
				$sCurrentRowValue = Strip( substr( $sCurrentRowKeyValue, $sEqualPos+1 ) );
					
				# Ensure, that there is a valid key name
					
				if( $sCurrentRowKey eq "" ) {
					return( -3 );
				}
					
				# Get the number of backslashes at the end of row
					
				$sBackSlashesNoOf = 0;
					
				for( $I = length( $sCurrentRowValue ); $I > 0; $I-- ) {
					if( substr( $sCurrentRowValue, $I-1, 1 ) eq '\\' ) {
						$sBackSlashesNoOf++;
					}
					else {
						last;
					}
				}
					
				# Process the indicator for a following row (number of backslashes = 1,3,5,...)
				
				if( $sBackSlashesNoOf % 2 != 0 ) {
					$sCurrentRowValue = Strip( substr( $sCurrentRowValue, 0, length( $sCurrentRowValue ) - 1 ) );
					$sKey4FollowRow   = $sCurrentRowKey;
					$sRowNo2Follow    = $sRowNo;
				}
					
				# Write key/value into the hash (replace duplicated backslashes with single once)
					
				$sCurrentRowValue = Replace( '\\\\\\\\', '\\\\', $sCurrentRowValue );
				$$rhKeyValueRows{$sCurrentRowKey} = $sCurrentRowValue;
				next;
			}
				
			# Otherwise (row is treated as a value row),...
				
			else {
				
				# Ensure, that a following row is wanted
					
				if( ( $sKey4FollowRow eq "" ) || ( ( $sRowNo2Follow + 1 ) != $sRowNo ) ) {
					return( -4 );
				}
					
				# Split row into key- and value- token
					
				$sCurrentRowKey   = "";
				$sCurrentRowValue = $sCurrentRowKeyValue;
					
				# Get number of backslashes
					
				$sBackSlashesNoOf = 0;
					
				for( $I = length( $sCurrentRowValue ); $I > 0; $I-- ) {
					if( substr( $sCurrentRowValue, $I-1, 1 ) eq '\\' ) {
						$sBackSlashesNoOf++;
					}
					else {
						last;
					}
				}
					
				# Process the indication for following row (number of backslashes = 1,3,5,...)
					
				if( $sBackSlashesNoOf % 2 != 0 ) {
					$sCurrentRowValue = Strip( substr( $sCurrentRowValue, 0, length( $sCurrentRowValue ) - 1 ) );
					$sRowNo2Follow    = $sRowNo;
				}
				else {
					$sRowNo2Follow = -1;
				}

				# Write key/value into hash (replace duplicated bashslashes with single once)
					
				$sCurrentRowValue = Replace( '\\\\\\\\', '\\\\', $sCurrentRowValue );
				$$rhKeyValueRows{$sKey4FollowRow} = $$rhKeyValueRows{$sKey4FollowRow}.$sCurrentRowValue;
				
				next;
			}
		}
			
		# Otherwise (empty row),...
			
		else {
			$sRowNo2Follow = -1;
			next;
		}
	}



	# Close file when needed and return the key/value pairs

	if( defined( $oFh ) ) {
		close( $oFh );
	}

	return( $rhKeyValueRows );
} 





###############################################################################
##                                                                           ##
##   P A R S E A R G U M E N T S                                             ##
##                                                                           ##
###############################################################################
sub ParseArguments($$$) {

	# Presettings

	my( $raArguments, $raMarkerDefinitions, $rhTargetHash ) = @_;
	my( $I, $sIndexPos, $sDummy, @aDummy, %hArgTokenIndex );



	# Check and extent the given list of parameter markers
		
	for( $I = 0; $I < scalar@$raMarkerDefinitions; $I++ ) {
		
		# Check the definition of the parameter markers
		
		@aDummy = split( ":", $$raMarkerDefinitions[$I]->{"Definition"} );
		foreach( @aDummy ) {
			$_ = Strip( uc( $_ ) );
		}
  
		# Check for (enough) definitions per parameter marker
		
		if( ( ! defined( $aDummy[0] ) ) || ( ! defined( $aDummy[1] ) ) ) {
			return( -1 );
		}
			
		# Check for valid type of parameter marker
		
		if( ( $aDummy[0] ne "FLAG" ) && ( $aDummy[0] ne "BOOLEAN" ) && ( $aDummy[0] ne "STRING" ) ) {
			return( -2 );
		}
			
		# Check for settings if parameter marker is a must in the list of arguments,...
			
		if( $aDummy[1] eq "Y" ) {
			$aDummy[1] = TRUE;
		}
		elsif( $aDummy[1] eq "N" ) {
			$aDummy[1] = FALSE;
		}
		elsif( $aDummy[0] eq "FLAG" ) {
			$aDummy[1] = FALSE;
		}
		else {
			return( -3 );
		}

		# Extent the list of parameter markers
		
		$$raMarkerDefinitions[$I]->{"FoundCounter"}    = 0;
		$$raMarkerDefinitions[$I]->{"Def.Type"}        = $aDummy[0];
		$$raMarkerDefinitions[$I]->{"Def.MustBeThere"} = $aDummy[1];
			
		# Additional Settings
			
		$hArgTokenIndex{$$raMarkerDefinitions[$I]->{"ArgToken"}} = $I;
	}



	# Walk thrue the given list of arguments
		
	for( $I = 0; $I < scalar@$raArguments; $I++ ) {
		
		# If argument matches one of the parameter markers,...
			
		if( defined( $hArgTokenIndex{$$raArguments[$I]} ) ) {
			
			# Presettings
				
			$sIndexPos = $hArgTokenIndex{$$raArguments[$I]};
			$$raMarkerDefinitions[$sIndexPos]->{"FoundCounter"}++;
   
			# Do processing depending on the defined type for the parameter marker
			
			if( $$raMarkerDefinitions[$sIndexPos]->{"Def.Type"} eq "FLAG" ) {
				$$raMarkerDefinitions[$sIndexPos]->{"Value"} = TRUE;
			}
			elsif( $$raMarkerDefinitions[$sIndexPos]->{"Def.Type"} eq "BOOLEAN" ) {
					
				# Ensure, that a arguments follows which is not a parameter marker
					
				if( ( $I + 1 ) < scalar@$raArguments ) {
					$sDummy = $hArgTokenIndex{$$raArguments[$I+1]};
					if( defined( $sDummy ) ) {
						return( -6 );
					}
				}
				else {
					return( -5 );
				}
					
				# Process the parameter marker following argument
				
				$sDummy = uc( $$raArguments[$I+1] );
				if( ( $sDummy eq "ON" )  || ( $sDummy eq "YES" ) || ( $sDummy eq "TRUE" ) ) {
					$$raMarkerDefinitions[$sIndexPos]->{"Value"} = TRUE;
				}
				elsif( ( $sDummy eq "OFF" ) || ( $sDummy eq "NO" )  || ( $sDummy eq "FALSE" ) ) {
					$$raMarkerDefinitions[$sIndexPos]->{"Value"} = FALSE;
				}
				else {
					return( -7 );
				}
				
				# Increase the argument index (to skip the second argument)
					
				$I++;
			}
			elsif( $$raMarkerDefinitions[$sIndexPos]->{"Def.Type"} eq "STRING" ) {
			
				# Ensure, that a arguments follows which is not a parameter marker
					
				if( ( $I + 1 ) < scalar@$raArguments ) {
					$sDummy = $hArgTokenIndex{$$raArguments[$I+1]};
					if( defined( $sDummy ) ) {
						return( -8 );
					}
				}
				else {
					return( -9 );
				}
					
				# Process the parameter marker following argument
					
				$$raMarkerDefinitions[$sIndexPos]->{"Value"} = $$raArguments[$I+1];
					
				# Increase the argument index (to skip the second argument)
					
				$I++;
			}
			else {
				return( -4 );
			}
		}
		else {
			return( -10 );
		}
	}



	# Ensure, that all "must" parameter markers are found in argument list
		
	for( $I = 0; $I < scalar@$raMarkerDefinitions; $I++ ) {
		
		# Check if every "must" parameter marker was found
			
		if( ( $$raMarkerDefinitions[$I]->{"Def.MustBeThere"} ) &&
			( $$raMarkerDefinitions[$I]->{"FoundCounter"} <= 0 ) ) {
			return( -11 );
		}
			
		# Check if a "must" parameter marker was found more than one time
			
		if( $$raMarkerDefinitions[$I]->{"FoundCounter"} > 1 ) {
			return( -12 );
		}
	}



	# Write the values/defaults for the defined parameter markers
		
	for( $I = 0; $I < scalar@$raMarkerDefinitions; $I++ ) {
			
		# If parameter marker was found in arguments,...
		
		if( $$raMarkerDefinitions[$I]->{"FoundCounter"} > 0 ) {
			$$rhTargetHash{$$raMarkerDefinitions[$I]->{"HashKey"}} = $$raMarkerDefinitions[$I]->{"Value"};
		}
			
		# If parameter marker wasn't found in arguments (but has a default),...
			
		elsif( exists( $$raMarkerDefinitions[$I]->{"Default"} ) ) {
			$$rhTargetHash{$$raMarkerDefinitions[$I]->{"HashKey"}} = $$raMarkerDefinitions[$I]->{"Default"};
		}
	}    


	
	return( OK );
}





###############################################################################
##                                                                           ##
##   R E P L A C E                                                           ##
##                                                                           ##
###############################################################################
sub Replace($$@) {

	my( $sFromPattern, $sToPattern, @aRc ) = @_;
	my $sStmt;

	for( @aRc ) {
		$sStmt="s/$sFromPattern/$sToPattern/g;";
		eval $sStmt;  
	}

	return wantarray ? @aRc : $aRc[0]; 
}





###############################################################################
##                                                                           ##
##   S P A C E                                                               ##
##                                                                           ##
###############################################################################
sub Space($$@) {

	my( $sNoOfSpaces, $sReplaceChar, @aRc ) = @_;
	my $sReplaceToken = ( ( defined( $sReplaceChar ) ) ? $sReplaceChar : " " ) x $sNoOfSpaces;

	for( @aRc ) {
		s/\s+/$sReplaceToken/g;
	}

	return wantarray ? @aRc : $aRc[0];
}





###############################################################################
##                                                                           ##
##   S T R I P                                                               ##
##                                                                           ##
###############################################################################
sub Strip(@) {

	my @aArray = @_;

	foreach( @aArray ) { 
		s/^\s+//;
		s/\s+$//;
	}

	return wantarray ? @aArray : $aArray[0];
}





###############################################################################
##                                                                           ##
##   T R A N S L A T E                                                       ##
##                                                                           ##
###############################################################################
sub Translate($$$@) {

	my( $sFromPattern, $sToPattern, $sOption, @aRc ) = @_;
	my $sStmt;

	if( !defined( $sOption ) ) {
		$sOption = "";
	}

	for( @aRc ) {
		$sStmt = "tr/$sFromPattern/$sToPattern/$sOption;";
		eval $sStmt;
	}

	return wantarray ? @aRc : $aRc[0];
}




